/* Blue Midnight Wish hash function as it will be submitted to 
   NIST for the second round of SHA-3 competition.
   Changes in this version (compared to the version submitted for the Round 1):

     1. Corrected IV for 224 and 384 version.
     2. Tweaked f0() and f1()
     3. Use of final compression function invocation.

   $Id: BlueMidnightWish_opt.c 69 2009-09-14 06:42:16Z joern $
*/													

/* 
   This is an optimized C code for Blue Midnight Wish hash function 
   as it will be submitted to NIST for the second round of SHA-3 competition.

   Programmed by Danilo Gligoroski, August 2009.
*/


#include <string.h> 
#include "SHA3api_ref.h"
#include "byteorder.h"

#define rotl32(x,n)   (((x) << n) | ((x) >> (32 - n)))
#define rotr32(x,n)   (((x) >> n) | ((x) << (32 - n)))

#define rotl64(x,n)   (((x) << n) | ((x) >> (64 - n)))
#define rotr64(x,n)   (((x) >> n) | ((x) << (64 - n)))

#define shl(x,n)      ((x) << n)
#define shr(x,n)      ((x) >> n)

/* BlueMidnightWish224 initial double chaining pipe */
const uint32_t i224p2[16] = {
	0x00010203ul, 0x04050607ul, 0x08090a0bul, 0x0c0d0e0ful,
	0x10111213ul, 0x14151617ul, 0x18191a1bul, 0x1c1d1e1ful,
	0x20212223ul, 0x24252627ul, 0x28292a2bul, 0x2c2d2e2ful,
	0x30313233ul, 0x34353637ul, 0x38393a3bul, 0x3c3d3e3ful,
};

/* BlueMidnightWish256 initial double chaining pipe */
const uint32_t i256p2[16] = {
	0x40414243ul, 0x44454647ul, 0x48494a4bul, 0x4c4d4e4ful,
	0x50515253ul, 0x54555657ul, 0x58595a5bul, 0x5c5d5e5ful,
	0x60616263ul, 0x64656667ul, 0x68696a6bul, 0x6c6d6e6ful,
	0x70717273ul, 0x74757677ul, 0x78797a7bul, 0x7c7d7e7ful,
};

/* BlueMidnightWish384 initial double chaining pipe */
const uint64_t i384p2[16] = {
	0x0001020304050607ull, 0x08090a0b0c0d0e0full,
	0x1011121314151617ull, 0x18191a1b1c1d1e1full,
	0x2021222324252627ull, 0x28292a2b2c2d2e2full,
	0x3031323334353637ull, 0x38393a3b3c3d3e3full,
	0x4041424344454647ull, 0x48494a4b4c4d4e4full,
	0x5051525354555657ull, 0x58595a5b5c5d5e5full,
	0x6061626364656667ull, 0x68696a6b6c6d6e6full,
	0x7071727374757677ull, 0x78797a7b7c7d7e7full
};

/* BlueMidnightWish512 initial double chaining pipe */
const uint64_t i512p2[16] = {
	0x8081828384858687ull, 0x88898a8b8c8d8e8full,
	0x9091929394959697ull, 0x98999a9b9c9d9e9full,
	0xa0a1a2a3a4a5a6a7ull, 0xa8a9aaabacadaeafull,
	0xb0b1b2b3b4b5b6b7ull, 0xb8b9babbbcbdbebfull,
	0xc0c1c2c3c4c5c6c7ull, 0xc8c9cacbcccdcecfull,
	0xd0d1d2d3d4d5d6d7ull, 0xd8d9dadbdcdddedfull,
	0xe0e1e2e3e4e5e6e7ull, 0xe8e9eaebecedeeefull,
	0xf0f1f2f3f4f5f6f7ull, 0xf8f9fafbfcfdfeffull
};

#define hashState224(x)  ((x)->pipe->p256)
#define hashState256(x)  ((x)->pipe->p256)
#define hashState384(x)  ((x)->pipe->p512)
#define hashState512(x)  ((x)->pipe->p512)

/* Table 1.3 logic functions for the 384 and 512 bit versions (left column) */
#define s32_0(x) (shr((x), 1) ^ shl((x), 3) ^ rotl32((x),  4) ^ rotl32((x), 19))
#define s32_1(x) (shr((x), 1) ^ shl((x), 2) ^ rotl32((x),  8) ^ rotl32((x), 23))
#define s32_2(x) (shr((x), 2) ^ shl((x), 1) ^ rotl32((x), 12) ^ rotl32((x), 25))
#define s32_3(x) (shr((x), 2) ^ shl((x), 2) ^ rotl32((x), 15) ^ rotl32((x), 29))
#define s32_4(x) (shr((x), 1) ^ (x))
#define s32_5(x) (shr((x), 2) ^ (x))
#define r32_1(x) rotl32((x),  3)
#define r32_2(x) rotl32((x),  7)
#define r32_3(x) rotl32((x), 13)
#define r32_4(x) rotl32((x), 16)
#define r32_5(x) rotl32((x), 19)
#define r32_6(x) rotl32((x), 23)
#define r32_7(x) rotl32((x), 27)


#define Compression256() \
{ \
	uint32_t p_00, p_01, p_02, p_03, p_04, p_05, p_06, p_07, p_08, p_09, \
		p_10, p_11, p_12, p_13, p_14, p_15, p_16, p_17, p_18, p_19, p_20, \
		p_21, p_22, p_23, p_24, p_25, p_26, p_27, p_28, p_29, p_30, p_31; \
	uint32_t te, to, xl, xh; \
\
	/* Mix the message block with the previous double pipe */ \
	p_00 = p256[ 0] ^ d32( 0); \
	p_01 = p256[ 1] ^ d32( 1); \
	p_02 = p256[ 2] ^ d32( 2); \
	p_03 = p256[ 3] ^ d32( 3); \
	p_04 = p256[ 4] ^ d32( 4); \
	p_05 = p256[ 5] ^ d32( 5); \
	p_06 = p256[ 6] ^ d32( 6); \
	p_07 = p256[ 7] ^ d32( 7); \
	p_08 = p256[ 8] ^ d32( 8); \
	p_09 = p256[ 9] ^ d32( 9); \
	p_10 = p256[10] ^ d32(10); \
	p_11 = p256[11] ^ d32(11); \
	p_12 = p256[12] ^ d32(12); \
	p_13 = p256[13] ^ d32(13); \
	p_14 = p256[14] ^ d32(14); \
	p_15 = p256[15] ^ d32(15); \
\
	/* These variables are used for the tweak */ \
	td32_00 = rotl32(d32( 0), 1); \
	td32_01 = rotl32(d32( 1), 2); \
	td32_02 = rotl32(d32( 2), 3); \
	td32_03 = rotl32(d32( 3), 4); \
	td32_04 = rotl32(d32( 4), 5); \
	td32_05 = rotl32(d32( 5), 6); \
	td32_06 = rotl32(d32( 6), 7); \
	td32_07 = rotl32(d32( 7), 8); \
	td32_08 = rotl32(d32( 8), 9); \
	td32_09 = rotl32(d32( 9),10); \
	td32_10 = rotl32(d32(10),11); \
	td32_11 = rotl32(d32(11),12); \
	td32_12 = rotl32(d32(12),13); \
	td32_13 = rotl32(d32(13),14); \
	td32_14 = rotl32(d32(14),15); \
	td32_15 = rotl32(d32(15),16); \
\
	{ \
		uint32_t t, t_16, t_17, t_18, t_19, t_20, t_21; \
\
		t_16 = p_01 - p_14; \
		t_17 = p_15 - p_12; \
		t_18 = p_14 - p_07; \
		t_19 = p_13 - p_06; \
		t_20 = p_08 - p_05; \
		t_21 = p_08 - p_01; \
\
		t = p_05 + t_18 + p_10 + p_13; \
		p_16 = s32_0(t) + p256[ 1]; \
		t = p_06 - p_08 + p_11 + p_14 - p_15; \
		p_17 = s32_1(t) + p256[ 2]; \
		t = p_00 + p_07 + p_09 + t_17; \
		p_18 = s32_2(t) + p256[ 3]; \
		t = p_00 + t_21 - p_10 + p_13; \
		p_19 = s32_3(t) + p256[ 4]; \
		t = t_16+p_02+p_09 - p_11; \
		p_20 = s32_4(t) + p256[ 5]; \
		t = p_03 - p_02 + p_10 + t_17; \
		p_21 = s32_0(t) + p256[ 6]; \
		t = p_04 - p_00 - p_03 - p_11 + p_13; \
		p_22 = s32_1(t) + p256[ 7]; \
		t = t_16 - p_04 - p_05 - p_12; \
		p_23 = s32_2(t) + p256[ 8]; \
		t = p_02 - p_05+t_19 - p_15; \
		p_24 = s32_3(t) + p256[ 9]; \
		t = p_00 - p_03 + p_06 + t_18; \
		p_25 = s32_4(t) + p256[10]; \
		t = t_21 - p_04 - p_07 + p_15; \
		p_26 = s32_0(t) + p256[11]; \
		t = t_20 - p_00 - p_02 + p_09; \
		p_27 = s32_1(t) + p256[12]; \
		t = p_01 + p_03 - p_06 - p_09 + p_10; \
		p_28 = s32_2(t) + p256[13]; \
		t = p_02 + p_04 + p_07 + p_10 + p_11; \
		p_29 = s32_3(t) + p256[14]; \
		t = p_03 + t_20 - p_11 - p_12; \
		p_14 = s32_4(t) + p256[15]; \
		t = p_12 - p_04 + t_19 - p_09; \
		p_15 = s32_0(t) + p256[ 0]; \
	} \
\
	p_00 = p_16; \
	p_01 = p_17; \
	p_02 = p_18; \
	p_03 = p_19; \
	p_04 = p_20; \
	p_05 = p_21; \
	p_06 = p_22; \
	p_07 = p_23; \
	p_08 = p_24; \
	p_09 = p_25; \
	p_10 = p_26; \
	p_11 = p_27; \
	p_12 = p_28; \
	p_13 = p_29; \
\
	/*** The 16-round message expansion part ***/ \
\
	xl = p_16 = s32_1(p_00) + s32_2(p_01) + s32_3(p_02) \
		+ s32_0(p_03) + s32_1(p_04) + s32_2(p_05) + s32_3(p_06) \
		+ s32_0(p_07) + s32_1(p_08) + s32_2(p_09) + s32_3(p_10) \
		+ s32_0(p_11) + s32_1(p_12) + s32_2(p_13) + s32_3(p_14) \
		+ s32_0(p_15) + ((td32_00 + td32_03 - td32_10 + 0x55555550ul) \
		^ p256[ 7]); \
	p_17 = s32_1(p_01) + s32_2(p_02) + s32_3(p_03) \
		+ s32_0(p_04) + s32_1(p_05) + s32_2(p_06) + s32_3(p_07) \
		+ s32_0(p_08) + s32_1(p_09) + s32_2(p_10) + s32_3(p_11) \
		+ s32_0(p_12) + s32_1(p_13) + s32_2(p_14) + s32_3(p_15) \
		+ s32_0(p_16) + ((td32_01 + td32_04 - td32_11 + 0x5aaaaaa5ul) \
		^ p256[ 8]); \
	xl ^= p_17; \
	te = p_14 + p_12 + p_10 + p_08 + p_06 + p_04 + p_02; \
	to = p_15 + p_13 + p_11 + p_09 + p_07 + p_05 + p_03; \
\
	/* expand32_22(18); */ \
	p_18 = te + r32_1(p_03) + r32_2(p_05) + r32_3(p_07) \
		+ r32_4(p_09) + r32_5(p_11) + r32_6(p_13) + r32_7(p_15) \
		+ s32_4(p_16) + s32_5(p_17) + ((td32_02 + td32_05 - td32_12 \
		+ 0x5ffffffaul) ^ p256[ 9]); \
	xl ^= p_18; \
	/* expand32_21(19); */ \
	p_19 = to + r32_1(p_04) + r32_2(p_06) + r32_3(p_08) \
		+ r32_4(p_10) + r32_5(p_12) + r32_6(p_14) + r32_7(p_16) \
		+ s32_4(p_17) + s32_5(p_18) + ((td32_03 + td32_06 - td32_13 \
		+ 0x6555554ful) ^ p256[10]); \
	xl ^= p_19; \
	te += p_16 - p_02; \
	/* expand32_22(20); */ \
	p_20 = te + r32_1(p_05) + r32_2(p_07) + r32_3(p_09) \
		+ r32_4(p_11) + r32_5(p_13) + r32_6(p_15) + r32_7(p_17) \
		+ s32_4(p_18) + s32_5( p_19) + ((td32_04 + td32_07 - td32_14 \
		+ 0x6aaaaaa4ul) ^ p256[11]); \
	xl ^= p_20; \
	to = to + p_17 - p_03; \
	/* expand32_21(21); */ \
	p_21 = to + r32_1(p_06) + r32_2(p_08) + r32_3(p_10) \
		+ r32_4(p_12) + r32_5(p_14) + r32_6(p_16) + r32_7(p_18) \
		+ s32_4(p_19) + s32_5( p_20) + ((td32_05 + td32_08 - td32_15 \
		+ 0x6ffffff9ul) ^ p256[12]); \
	xl ^= p_21; \
	te += p_18 - p_04; \
	/* expand32_22(22); */ \
	p_22 = te + r32_1(p_07) + r32_2(p_09) + r32_3(p_11) \
		+ r32_4(p_13) + r32_5(p_15) + r32_6(p_17) + r32_7(p_19) \
		+ s32_4(p_20) + s32_5(p_21) + ((td32_06 + td32_09 - td32_00 \
		+ 0x7555554eul) ^ p256[13]); \
	xl ^= p_22; \
	to += p_19 - p_05; \
	/* expand32_21(23); */ \
	p_23 = to + r32_1(p_08) + r32_2(p_10) + r32_3(p_12) \
		+ r32_4(p_14) + r32_5(p_16) + r32_6(p_18) + r32_7(p_20) \
		+ s32_4(p_21) + s32_5(p_22) + ((td32_07 + td32_10 - td32_01 \
		+ 0x7aaaaaa3ul) ^ p256[14]); \
	xl ^= p_23; \
	te += p_20 - p_06; \
	/* expand32_22(24); */ \
	p_24 = te + r32_1(p_09) + r32_2(p_11) + r32_3(p_13) \
		+ r32_4(p_15) + r32_5(p_17) + r32_6(p_19) + r32_7(p_21) \
		+ s32_4(p_22) + s32_5(p_23) + ((td32_08 + td32_11 - td32_02 \
		+ 0x7ffffff8ul) ^ p256[15]); \
	xh = xl ^ p_24; \
	to += p_21 - p_07; \
	/* expand32_21(25); */ \
	p_25 = to + r32_1(p_10) + r32_2(p_12) + r32_3(p_14) \
		+ r32_4(p_16) + r32_5(p_18) + r32_6(p_20) + r32_7(p_22) \
		+ s32_4(p_23) + s32_5(p_24) + ((td32_09 + td32_12 - td32_03 \
		+ 0x8555554dul) ^ p256[ 0]); \
	xh ^= p_25; \
	te += p_22 - p_08; \
	/* expand32_22(26); */ \
	p_26 = te + r32_1(p_11) + r32_2(p_13) + r32_3(p_15) \
		+ r32_4(p_17) + r32_5(p_19) + r32_6(p_21) + r32_7(p_23) \
		+ s32_4(p_24) + s32_5(p_25) + ((td32_10 + td32_13 - td32_04 \
		+ 0x8aaaaaa2ul) ^ p256[ 1]); \
	xh ^= p_26; \
	to += p_23 - p_09; \
	/* expand32_21(27); */ \
	p_27 = to + r32_1(p_12) + r32_2(p_14) + r32_3(p_16) \
		+ r32_4(p_18) + r32_5(p_20) + r32_6(p_22) + r32_7(p_24) \
		+ s32_4(p_25) + s32_5(p_26) + ((td32_11 + td32_14 - td32_05 \
		+ 0x8ffffff7ul) ^ p256[ 2]); \
	xh ^= p_27; \
	te += p_24 - p_10; \
	/* expand32_22(28); */ \
	p_28 = te + r32_1(p_13) + r32_2(p_15) + r32_3(p_17) \
		+ r32_4(p_19) + r32_5(p_21) + r32_6(p_23) + r32_7(p_25) \
		+ s32_4(p_26) + s32_5(p_27) + ((td32_12 + td32_15 - td32_06 \
		+ 0x9555554cul) ^ p256[ 3]); \
	xh ^= p_28; \
	to += p_25 - p_11; \
	/* expand32_21(29); */ \
	p_29 = to + r32_1(p_14) + r32_2(p_16) + r32_3(p_18) \
		+ r32_4(p_20) + r32_5(p_22) + r32_6(p_24) + r32_7(p_26) \
		+ s32_4(p_27) + s32_5(p_28) + ((td32_13 + td32_00 - td32_07 \
		+ 0x9aaaaaa1ul) ^ p256[ 4]); \
	xh ^= p_29; \
	te += p_26 - p_12; \
	/* expand32_22(30); */ \
	p_30 = te + r32_1(p_15) + r32_2(p_17) + r32_3(p_19) \
		+ r32_4(p_21) + r32_5(p_23) + r32_6(p_25) + r32_7(p_27) \
		+ s32_4(p_28) + s32_5(p_29) + ((td32_14 + td32_01 - td32_08 \
		+ 0x9ffffff6ul) ^ p256[ 5]); \
	xh ^= p_30; \
	to += p_27 - p_13; \
	/* expand32_21(31); */ \
	p_31 = to + r32_1(p_16) + r32_2(p_18) + r32_3(p_20) \
		+ r32_4(p_22) + r32_5(p_24) + r32_6(p_26) + r32_7(p_28) \
		+ s32_4(p_29) + s32_5(p_30) + ((td32_15 + td32_02 - td32_09 \
		+ 0xa555554bul) ^ p256[ 6]); \
	xh ^= p_31; \
\
	/* Compute the double chaining pipe for the next message block */ \
	p256[ 0] = (shl(xh, 5) ^ shr(p_16,5) ^ d32( 0)) + (xl ^ p_24 ^ p_00); \
	p256[ 1] = (shr(xh, 7) ^ shl(p_17,8) ^ d32( 1)) + (xl ^ p_25 ^ p_01); \
	p256[ 2] = (shr(xh, 5) ^ shl(p_18,5) ^ d32( 2)) + (xl ^ p_26 ^ p_02); \
	p256[ 3] = (shr(xh, 1) ^ shl(p_19,5) ^ d32( 3)) + (xl ^ p_27 ^ p_03); \
	p256[ 4] = (shr(xh, 3) ^ p_20 ^ d32( 4)) + (xl ^ p_28 ^ p_04); \
	p256[ 5] = (shl(xh, 6) ^ shr(p_21,6) ^ d32( 5)) + (xl ^ p_29 ^ p_05); \
	p256[ 6] = (shr(xh, 4) ^ shl(p_22,6) ^ d32( 6)) + (xl ^ p_30 ^ p_06); \
	p256[ 7] = (shr(xh,11) ^ shl(p_23,2) ^ d32( 7)) + (xl ^ p_31 ^ p_07); \
\
	p256[ 8] = rotl32(p256[4], 9) + (xh ^ p_24 ^ d32( 8)) \
		+ (shl(xl,8) ^ p_23 ^ p_08); \
	p256[ 9] = rotl32(p256[5],10) + (xh ^ p_25 ^ d32( 9)) \
		+ (shr(xl,6) ^ p_16 ^ p_09); \
	p256[10] = rotl32(p256[6],11) + (xh ^ p_26 ^ d32(10)) \
		+ (shl(xl,6) ^ p_17 ^ p_10); \
	p256[11] = rotl32(p256[7],12) + (xh ^ p_27 ^ d32(11)) \
		+ (shl(xl,4) ^ p_18 ^ p_11); \
	p256[12] = rotl32(p256[0],13) + (xh ^ p_28 ^ d32(12)) \
		+ (shr(xl,3) ^ p_19 ^ p_12); \
	p256[13] = rotl32(p256[1],14) + (xh ^ p_29 ^ d32(13)) \
		+ (shr(xl,4) ^ p_20 ^ p_13); \
	p256[14] = rotl32(p256[2],15) + (xh ^ p_30 ^ d32(14)) \
		+ (shr(xl,7) ^ p_21 ^ p_14); \
	p256[15] = rotl32(p256[3],16) + (xh ^ p_31 ^ d32(15)) \
		+ (shr(xl,2) ^ p_22 ^ p_15); \
}


#define FinalCompression256() \
{ \
	uint32_t p_00, p_01, p_02, p_03, p_04, p_05, p_06, p_07, p_08, p_09, \
		p_10, p_11, p_12, p_13, p_14, p_15, p_16, p_17, p_18, p_19, p_20, \
		p_21, p_22, p_23, p_24, p_25, p_26, p_27, p_28, p_29, p_30, p_31; \
	uint32_t te, to, xl, xh; \
\
	/* Mix the final double pipe with a constant */ \
	p_00 = p256[ 0] ^ 0xaaaaaaa0ul; \
	p_01 = p256[ 1] ^ 0xaaaaaaa1ul; \
	p_02 = p256[ 2] ^ 0xaaaaaaa2ul; \
	p_03 = p256[ 3] ^ 0xaaaaaaa3ul; \
	p_04 = p256[ 4] ^ 0xaaaaaaa4ul; \
	p_05 = p256[ 5] ^ 0xaaaaaaa5ul; \
	p_06 = p256[ 6] ^ 0xaaaaaaa6ul; \
	p_07 = p256[ 7] ^ 0xaaaaaaa7ul; \
	p_08 = p256[ 8] ^ 0xaaaaaaa8ul; \
	p_09 = p256[ 9] ^ 0xaaaaaaa9ul; \
	p_10 = p256[10] ^ 0xaaaaaaaaul; \
	p_11 = p256[11] ^ 0xaaaaaaabul; \
	p_12 = p256[12] ^ 0xaaaaaaacul; \
	p_13 = p256[13] ^ 0xaaaaaaadul; \
	p_14 = p256[14] ^ 0xaaaaaaaeul; \
	p_15 = p256[15] ^ 0xaaaaaaaful; \
\
	/* These variables are used for the tweak */ \
	td32_00 = rotl32(p256[ 0], 1); \
	td32_01 = rotl32(p256[ 1], 2); \
	td32_02 = rotl32(p256[ 2], 3); \
	td32_03 = rotl32(p256[ 3], 4); \
	td32_04 = rotl32(p256[ 4], 5); \
	td32_05 = rotl32(p256[ 5], 6); \
	td32_06 = rotl32(p256[ 6], 7); \
	td32_07 = rotl32(p256[ 7], 8); \
	td32_08 = rotl32(p256[ 8], 9); \
	td32_09 = rotl32(p256[ 9],10); \
	td32_10 = rotl32(p256[10],11); \
	td32_11 = rotl32(p256[11],12); \
	td32_12 = rotl32(p256[12],13); \
	td32_13 = rotl32(p256[13],14); \
	td32_14 = rotl32(p256[14],15); \
	td32_15 = rotl32(p256[15],16); \
\
	{ \
		uint32_t t, t_16, t_17, t_18, t_19, t_20, t_21; \
\
		t_16 = p_01 - p_14; \
		t_17 = p_15 - p_12; \
		t_18 = p_14 - p_07; \
		t_19 = p_13 - p_06; \
		t_20 = p_08 - p_05; \
		t_21 = p_08 - p_01; \
\
		t = (p_05 + t_18 + p_10 + p_13); \
		p_16 = s32_0(t) + 0xaaaaaaa1ul; \
		t = (p_06 - p_08 + p_11 + p_14 - p_15); \
		p_17 = s32_1(t) + 0xaaaaaaa2ul; \
		t = (p_00 + p_07 + p_09 + t_17); \
		p_18 = s32_2(t) + 0xaaaaaaa3ul; \
		t = (p_00 + t_21 - p_10 + p_13); \
		p_19 = s32_3(t) + 0xaaaaaaa4ul; \
		t = (t_16 + p_02 + p_09 - p_11); \
		p_20 = s32_4(t) + 0xaaaaaaa5ul; \
		t = (p_03 - p_02 + p_10 + t_17); \
		p_21 = s32_0(t) + 0xaaaaaaa6ul; \
		t = (p_04 - p_00 - p_03 - p_11 + p_13); \
		p_22 = s32_1(t) + 0xaaaaaaa7ul; \
		t = (t_16 - p_04 - p_05 - p_12); \
		p_23 = s32_2(t) + 0xaaaaaaa8ul; \
		t = (p_02 - p_05 + t_19 - p_15); \
		p_24 = s32_3(t) + 0xaaaaaaa9ul; \
		t = (p_00 - p_03 + p_06 + t_18); \
		p_25 = s32_4(t) + 0xaaaaaaaaul; \
		t = (t_21 - p_04 - p_07 + p_15); \
		p_26 = s32_0(t) + 0xaaaaaaabul; \
		t = (t_20 - p_00 - p_02 + p_09); \
		p_27 = s32_1(t) + 0xaaaaaaacul; \
		t = (p_01 + p_03 - p_06 - p_09 + p_10); \
		p_28 = s32_2(t) + 0xaaaaaaadul; \
		t = (p_02 + p_04 + p_07 + p_10 + p_11); \
		p_29 = s32_3(t) + 0xaaaaaaaeul; \
		p_30 = (p_03 + t_20 - p_11 - p_12); \
		p_14 = s32_4(p_30) + 0xaaaaaaaful; \
		p_31 = (p_12 - p_04 + t_19 - p_09); \
		p_15 = s32_0(p_31) + 0xaaaaaaa0ul; \
	} \
\
	p_00 = p_16; \
	p_01 = p_17; \
	p_02 = p_18; \
	p_03 = p_19; \
	p_04 = p_20; \
	p_05 = p_21; \
	p_06 = p_22; \
	p_07 = p_23; \
	p_08 = p_24; \
	p_09 = p_25; \
	p_10 = p_26; \
	p_11 = p_27; \
	p_12 = p_28; \
	p_13 = p_29; \
\
	/*** The 16-round message expansion part ***/ \
\
	xl = p_16 = s32_1(p_00) + s32_2(p_01) + s32_3(p_02) \
		+ s32_0(p_03) + s32_1(p_04) + s32_2(p_05) + s32_3(p_06) \
		+ s32_0(p_07) + s32_1(p_08) + s32_2(p_09) + s32_3(p_10) \
		+ s32_0(p_11) + s32_1(p_12) + s32_2(p_13) + s32_3(p_14) \
		+ s32_0(p_15) + ((td32_00 + td32_03 - td32_10 + 0x55555550ul) \
		^ 0xaaaaaaa7ul); \
	p_17 = s32_1(p_01) + s32_2(p_02) + s32_3(p_03) \
		+ s32_0(p_04) + s32_1(p_05) + s32_2(p_06) + s32_3(p_07) \
		+ s32_0(p_08) + s32_1(p_09) + s32_2(p_10) + s32_3(p_11) \
		+ s32_0(p_12) + s32_1(p_13) + s32_2(p_14) + s32_3(p_15) \
		+ s32_0(p_16) + ((td32_01 + td32_04 - td32_11 + 0x5aaaaaa5ul) \
		^ 0xaaaaaaa8ul); \
	xl ^= p_17; \
	te = p_14 + p_12 + p_10 + p_08 + p_06 + p_04 + p_02; \
	to = p_15 + p_13 + p_11 + p_09 + p_07 + p_05 + p_03; \
\
	/* expand32_22(18); */ \
	p_18 = te + r32_1(p_03) + r32_2(p_05) + r32_3(p_07) \
		+ r32_4(p_09) + r32_5(p_11) + r32_6(p_13) + r32_7(p_15) \
		+ s32_4(p_16) + s32_5(p_17) + ((td32_02 + td32_05 - td32_12 \
		+ 0x5ffffffaul) ^ 0xaaaaaaa9ul); \
	xl ^= p_18; \
	/* expand32_21(19); */ \
	p_19 = to + r32_1(p_04) + r32_2(p_06) + r32_3(p_08) \
		+ r32_4(p_10) + r32_5(p_12) + r32_6(p_14) + r32_7(p_16) \
		+ s32_4(p_17) + s32_5(p_18) + ((td32_03 + td32_06 - td32_13 \
		+ 0x6555554ful) ^ 0xaaaaaaaaul); \
	xl ^= p_19; \
	te += p_16 - p_02; \
	/* expand32_22(20); */ \
	p_20 = te + r32_1(p_05) + r32_2(p_07) + r32_3(p_09) \
		+ r32_4(p_11) + r32_5(p_13) + r32_6(p_15) + r32_7(p_17) \
		+ s32_4(p_18) + s32_5( p_19) + ((td32_04 + td32_07 - td32_14 \
		+ 0x6aaaaaa4ul) ^ 0xaaaaaaabul); \
	xl ^= p_20; \
	to = to + p_17 - p_03; \
	/* expand32_21(21); */ \
	p_21 = to + r32_1(p_06) + r32_2(p_08) + r32_3(p_10) \
		+ r32_4(p_12) + r32_5(p_14) + r32_6(p_16) + r32_7(p_18) \
		+ s32_4(p_19) + s32_5( p_20) + ((td32_05 + td32_08 - td32_15 \
		+ 0x6ffffff9ul) ^ 0xaaaaaaacul); \
	xl ^= p_21; \
	te += p_18 - p_04; \
	/* expand32_22(22); */ \
	p_22 = te + r32_1(p_07) + r32_2(p_09) + r32_3(p_11) \
		+ r32_4(p_13) + r32_5(p_15) + r32_6(p_17) + r32_7(p_19) \
		+ s32_4(p_20) + s32_5(p_21) + ((td32_06 + td32_09 - td32_00 \
		+ 0x7555554eul) ^ 0xaaaaaaadul); \
	xl ^= p_22; \
	to += p_19 - p_05; \
	/* expand32_21(23); */ \
	p_23 = to + r32_1(p_08) + r32_2(p_10) + r32_3(p_12) \
		+ r32_4(p_14) + r32_5(p_16) + r32_6(p_18) + r32_7(p_20) \
		+ s32_4(p_21) + s32_5(p_22) + ((td32_07 + td32_10 - td32_01 \
		+ 0x7aaaaaa3ul) ^ 0xaaaaaaaeul); \
	xl ^= p_23; \
	te += p_20 - p_06; \
	/* expand32_22(24); */ \
	p_24 = te + r32_1(p_09) + r32_2(p_11) + r32_3(p_13) \
		+ r32_4(p_15) + r32_5(p_17) + r32_6(p_19) + r32_7(p_21) \
		+ s32_4(p_22) + s32_5(p_23) + ((td32_08 + td32_11 - td32_02 \
		+ 0x7ffffff8ul) ^ 0xaaaaaaaful); \
	xh = xl ^ p_24; \
	to += p_21 - p_07; \
	/* expand32_21(25); */ \
	p_25 = to + r32_1(p_10) + r32_2(p_12) + r32_3(p_14) \
		+ r32_4(p_16) + r32_5(p_18) + r32_6(p_20) + r32_7(p_22) \
		+ s32_4(p_23) + s32_5(p_24) + ((td32_09 + td32_12 - td32_03 \
		+ 0x8555554dul) ^ 0xaaaaaaa0ul); \
	te += p_22 - p_08; \
	xh ^= p_25; \
	/* expand32_22(26); */ \
	p_26 = te + r32_1(p_11) + r32_2(p_13) + r32_3(p_15) \
		+ r32_4(p_17) + r32_5(p_19) + r32_6(p_21) + r32_7(p_23) \
		+ s32_4(p_24) + s32_5(p_25) + ((td32_10 + td32_13 - td32_04 \
		+ 0x8aaaaaa2ul) ^ 0xaaaaaaa1ul); \
	xh ^= p_26; \
	to += p_23 - p_09; \
	/* expand32_21(27); */ \
	p_27 = to + r32_1(p_12) + r32_2(p_14) + r32_3(p_16) \
		+ r32_4(p_18) + r32_5(p_20) + r32_6(p_22) + r32_7(p_24) \
		+ s32_4(p_25) + s32_5(p_26) + ((td32_11 + td32_14 - td32_05 \
		+ 0x8ffffff7ul) ^ 0xaaaaaaa2ul); \
	xh ^= p_27; \
	te += p_24 - p_10; \
	/* expand32_22(28); */ \
	p_28 = te + r32_1(p_13) + r32_2(p_15) + r32_3(p_17) \
		+ r32_4(p_19) + r32_5(p_21) + r32_6(p_23) + r32_7(p_25) \
		+ s32_4(p_26) + s32_5(p_27) + ((td32_12 + td32_15 - td32_06 \
		+ 0x9555554cul) ^ 0xaaaaaaa3ul); \
	xh ^= p_28; \
	to += p_25 - p_11; \
	/* expand32_21(29); */ \
	p_29 = to + r32_1(p_14) + r32_2(p_16) + r32_3(p_18) \
		+ r32_4(p_20) + r32_5(p_22) + r32_6(p_24) + r32_7(p_26) \
		+ s32_4(p_27) + s32_5(p_28) + ((td32_13 + td32_00 - td32_07 \
		+ 0x9aaaaaa1ul) ^ 0xaaaaaaa4ul); \
	xh ^= p_29; \
	te += p_26 - p_12; \
	/* expand32_22(30); */ \
	p_30 = te + r32_1(p_15) + r32_2(p_17) + r32_3(p_19) \
		+ r32_4(p_21) + r32_5(p_23) + r32_6(p_25) + r32_7(p_27) \
		+ s32_4(p_28) + s32_5(p_29) + ((td32_14 + td32_01 - td32_08 \
		+ 0x9ffffff6ul) ^ 0xaaaaaaa5ul); \
	xh ^= p_30; \
	to += p_27 - p_13; \
	/* expand32_21(31); */ \
	p_31 = to + r32_1(p_16) + r32_2(p_18) + r32_3(p_20) \
		+ r32_4(p_22) + r32_5(p_24) + r32_6(p_26) + r32_7(p_28) \
		+ s32_4(p_29) + s32_5(p_30) + ((td32_15 + td32_02 - td32_09 \
		+ 0xa555554bul) ^ 0xaaaaaaa6ul); \
	xh ^= p_31; \
\
	/* Compute the double chaining pipe for the final message */ \
	p256[ 0] = (shl(xh, 5) ^ shr(p_16,5) ^ p256[ 0]) + (xl ^ p_24 ^ p_00); \
	p256[ 1] = (shr(xh, 7) ^ shl(p_17,8) ^ p256[ 1]) + (xl ^ p_25 ^ p_01); \
	p256[ 2] = (shr(xh, 5) ^ shl(p_18,5) ^ p256[ 2]) + (xl ^ p_26 ^ p_02); \
	p256[ 3] = (shr(xh, 1) ^ shl(p_19,5) ^ p256[ 3]) + (xl ^ p_27 ^ p_03); \
	p256[ 4] = (shr(xh, 3) ^ p_20 ^ p256[ 4]) + (xl ^ p_28 ^ p_04); \
	p256[ 5] = (shl(xh, 6) ^ shr(p_21,6) ^ p256[ 5]) + (xl ^ p_29 ^ p_05); \
	p256[ 6] = (shr(xh, 4) ^ shl(p_22,6) ^ p256[ 6]) + (xl ^ p_30 ^ p_06); \
	p256[ 7] = (shr(xh,11) ^ shl(p_23,2) ^ p256[ 7]) + (xl ^ p_31 ^ p_07); \
\
	p256[ 8] = rotl32(p256[4], 9) + (xh ^ p_24 ^ p256[ 8]) \
		+ (shl(xl,8) ^ p_23 ^ p_08); \
	p256[ 9] = rotl32(p256[5],10) + (xh ^ p_25 ^ p256[ 9]) \
		+ (shr(xl,6) ^ p_16 ^ p_09); \
	p256[10] = rotl32(p256[6],11) + (xh ^ p_26 ^ p256[10]) \
		+ (shl(xl,6) ^ p_17 ^ p_10); \
	p256[11] = rotl32(p256[7],12) + (xh ^ p_27 ^ p256[11]) \
		+ (shl(xl,4) ^ p_18 ^ p_11); \
	p256[12] = rotl32(p256[0],13) + (xh ^ p_28 ^ p256[12]) \
		+ (shr(xl,3) ^ p_19 ^ p_12); \
	p256[13] = rotl32(p256[1],14) + (xh ^ p_29 ^ p256[13]) \
		+ (shr(xl,4) ^ p_20 ^ p_13); \
	p256[14] = rotl32(p256[2],15) + (xh ^ p_30 ^ p256[14]) \
		+ (shr(xl,7) ^ p_21 ^ p_14); \
	p256[15] = rotl32(p256[3],16) + (xh ^ p_31 ^ p256[15]) \
		+ (shr(xl,2) ^ p_22 ^ p_15); \
}


/* Table 1.3 logic functions for the 384 and 512 bit versions (right column) */
#define s64_0(x) (shr((x), 1) ^ shl((x), 3) ^ rotl64((x),  4) ^ rotl64((x), 37))
#define s64_1(x) (shr((x), 1) ^ shl((x), 2) ^ rotl64((x), 13) ^ rotl64((x), 43))
#define s64_2(x) (shr((x), 2) ^ shl((x), 1) ^ rotl64((x), 19) ^ rotl64((x), 53))
#define s64_3(x) (shr((x), 2) ^ shl((x), 2) ^ rotl64((x), 28) ^ rotl64((x), 59))
#define s64_4(x) (shr((x), 1) ^ (x))
#define s64_5(x) (shr((x), 2) ^ (x))
#define r64_1(x) rotl64((x),  5)
#define r64_2(x) rotl64((x), 11)
#define r64_3(x) rotl64((x), 27)
#define r64_4(x) rotl64((x), 32)
#define r64_5(x) rotl64((x), 37)
#define r64_6(x) rotl64((x), 43)
#define r64_7(x) rotl64((x), 53)

#define Compression512() \
{ \
	uint64_t p_00, p_01, p_02, p_03, p_04, p_05, p_06, p_07, p_08, p_09, \
		p_10, p_11, p_12, p_13, p_14, p_15, p_16, p_17, p_18, p_19, p_20, \
		p_21, p_22, p_23, p_24, p_25, p_26, p_27, p_28, p_29, p_30, p_31; \
	uint64_t te, to, xl, xh, *p = p512, *d = D64; \
\
	/* Mix the message block with the previous double pipe */ \
	p_00 = *p++ ^ *d++; \
	p_01 = *p++ ^ *d++; \
	p_02 = *p++ ^ *d++; \
	p_03 = *p++ ^ *d++; \
	p_04 = *p++ ^ *d++; \
	p_05 = *p++ ^ *d++; \
	p_06 = *p++ ^ *d++; \
	p_07 = *p++ ^ *d++; \
	p_08 = *p++ ^ *d++; \
	p_09 = *p++ ^ *d++; \
	p_10 = *p++ ^ *d++; \
	p_11 = *p++ ^ *d++; \
	p_12 = *p++ ^ *d++; \
	p_13 = *p++ ^ *d++; \
	p_14 = *p++ ^ *d++; \
	p_15 = *p++ ^ *d++; \
\
	/* These variables are used for the tweak */ \
	td64_00 = rotl64(d64( 0), 1); \
	td64_01 = rotl64(d64( 1), 2); \
	td64_02 = rotl64(d64( 2), 3); \
	td64_03 = rotl64(d64( 3), 4); \
	td64_04 = rotl64(d64( 4), 5); \
	td64_05 = rotl64(d64( 5), 6); \
	td64_06 = rotl64(d64( 6), 7); \
	td64_07 = rotl64(d64( 7), 8); \
	td64_08 = rotl64(d64( 8), 9); \
	td64_09 = rotl64(d64( 9),10); \
	td64_10 = rotl64(d64(10),11); \
	td64_11 = rotl64(d64(11),12); \
	td64_12 = rotl64(d64(12),13); \
	td64_13 = rotl64(d64(13),14); \
	td64_14 = rotl64(d64(14),15); \
	td64_15 = rotl64(d64(15),16); \
\
	{ \
		uint64_t t, t_16, t_17, t_18, t_19, t_20, t_21; \
\
		t_16 = p_01 - p_14; \
		t_17 = p_15 - p_12; \
		t_18 = p_14 - p_07; \
		t_19 = p_13 - p_06; \
		t_20 = p_08 - p_05; \
		t_21 = p_08 - p_01; \
\
		t = (p_05 + t_18 + p_10 + p_13); \
		p_16 = s64_0(t) + p512[ 1]; \
		t = (p_06 - p_08 + p_11 + p_14 - p_15); \
		p_17 = s64_1(t) + p512[ 2]; \
		t = (p_00 + p_07 + p_09 + t_17); \
		p_18 = s64_2(t) + p512[ 3]; \
		t = (p_00 + t_21 - p_10 + p_13); \
		p_19 = s64_3(t) + p512[ 4]; \
		t = (t_16 + p_02 + p_09 - p_11); \
		p_20 = s64_4(t) + p512[ 5]; \
		t = (p_03 - p_02 + p_10 + t_17); \
		p_21 = s64_0(t) + p512[ 6]; \
		t = (p_04 - p_00 - p_03 - p_11+p_13); \
		p_22 = s64_1(t) + p512[ 7]; \
		t = (t_16 - p_04 - p_05 - p_12); \
		p_23 = s64_2(t) + p512[ 8]; \
		t = (p_02 - p_05 + t_19 - p_15); \
		p_24 = s64_3(t) + p512[ 9]; \
		t = (p_00 - p_03 + p_06 + t_18); \
		p_25 = s64_4(t) + p512[10]; \
		t = (t_21 - p_04 - p_07 + p_15); \
		p_26 = s64_0(t) + p512[11]; \
		t = (t_20 - p_00 - p_02 + p_09); \
		p_27 = s64_1(t) + p512[12]; \
		t = (p_01 + p_03 - p_06 - p_09 + p_10); \
		p_28 = s64_2(t) + p512[13]; \
		t = (p_02 + p_04 + p_07 + p_10 + p_11); \
		p_29 = s64_3(t) + p512[14]; \
		p_30 = (p_03 + t_20 - p_11 - p_12); \
		p_14 = s64_4(p_30) + p512[15]; \
		p_31 = (p_12 - p_04 + t_19 - p_09); \
		p_15 = s64_0(p_31) + p512[ 0]; \
	} \
\
	p_00 = p_16; \
	p_01 = p_17; \
	p_02 = p_18; \
	p_03 = p_19; \
	p_04 = p_20; \
	p_05 = p_21; \
	p_06 = p_22; \
	p_07 = p_23; \
	p_08 = p_24; \
	p_09 = p_25; \
	p_10 = p_26; \
	p_11 = p_27; \
	p_12 = p_28; \
	p_13 = p_29; \
\
	/*** The 16-round message expansion part ***/ \
\
	xl = p_16 = s64_1(p_00) + s64_2(p_01) + s64_3(p_02) \
		+ s64_0(p_03) + s64_1(p_04) + s64_2(p_05) + s64_3(p_06) \
		+ s64_0(p_07) + s64_1(p_08) + s64_2(p_09) + s64_3(p_10) \
		+ s64_0(p_11) + s64_1(p_12) + s64_2(p_13) + s64_3(p_14) \
		+ s64_0(p_15) \
		+ ((td64_00 + td64_03 - td64_10 + 0x5555555555555550ull) ^ p512[ 7]); \
	p_17 = s64_1(p_01) + s64_2(p_02) + s64_3(p_03) \
		+ s64_0(p_04) + s64_1(p_05) + s64_2(p_06) + s64_3(p_07) \
		+ s64_0(p_08) + s64_1(p_09) + s64_2(p_10) + s64_3(p_11) \
		+ s64_0(p_12) + s64_1(p_13) + s64_2(p_14) + s64_3(p_15) \
		+ s64_0(p_16) \
		+ ((td64_01 + td64_04 - td64_11 + 0x5aaaaaaaaaaaaaa5ull) ^ p512[ 8]); \
	xl ^= p_17; \
	te = p_14 + p_12 + p_10 + p_08 + p_06 + p_04 + p_02; \
	to = p_15 + p_13 + p_11 + p_09 + p_07 + p_05 + p_03; \
\
	/* expand64_22(18); */ \
	p_18 = te + r64_1(p_03) + r64_2(p_05) + r64_3(p_07) \
		+ r64_4(p_09) + r64_5(p_11) + r64_6(p_13) + r64_7(p_15) \
		+ s64_4( p_16) + s64_5( p_17) \
		+ ((td64_02 + td64_05 - td64_12 + 0x5ffffffffffffffaull) ^ p512[ 9]); \
	xl ^= p_18; \
	/* expand64_21(19); */ \
	p_19 = to + r64_1(p_04) + r64_2(p_06) + r64_3(p_08) \
		+ r64_4(p_10) + r64_5(p_12) + r64_6(p_14) + r64_7(p_16) \
		+ s64_4( p_17) + s64_5( p_18) \
		+ ((td64_03 + td64_06 - td64_13 + 0x655555555555554full) ^ p512[10]); \
	xl ^= p_19; \
	te += p_16 - p_02; \
	/* expand64_22(20); */ \
	p_20 = te + r64_1(p_05) + r64_2(p_07) + r64_3(p_09) \
		+ r64_4(p_11) + r64_5(p_13) + r64_6(p_15) + r64_7(p_17) \
		+ s64_4( p_18) + s64_5( p_19) \
		+ ((td64_04 + td64_07 - td64_14 + 0x6aaaaaaaaaaaaaa4ull) ^ p512[11]); \
	xl ^= p_20; \
	to += p_17 - p_03; \
	/* expand64_21(21); */ \
	p_21 = to + r64_1(p_06) + r64_2(p_08) + r64_3(p_10) \
		+ r64_4(p_12) + r64_5(p_14) + r64_6(p_16) + r64_7(p_18) \
		+ s64_4( p_19) + s64_5( p_20) \
		+ ((td64_05 + td64_08 - td64_15 + 0x6ffffffffffffff9ull) ^ p512[12]); \
	xl ^= p_21; \
	te += p_18 - p_04; \
	/* expand64_22(22); */ \
	p_22  = te + r64_1(p_07) + r64_2(p_09) + r64_3(p_11) \
		+ r64_4(p_13) + r64_5(p_15) + r64_6(p_17) + r64_7(p_19) \
		+ s64_4( p_20) + s64_5( p_21)\
		+ ((td64_06 + td64_09 - td64_00 + 0x755555555555554eull) ^ p512[13]); \
	xl ^= p_22; \
	to += p_19 - p_05; \
	/* expand64_21(23); */ \
	p_23 = to + r64_1(p_08) + r64_2(p_10) + r64_3(p_12) \
		+ r64_4(p_14) + r64_5(p_16) + r64_6(p_18) + r64_7(p_20) \
		+ s64_4( p_21) + s64_5( p_22) \
		+ ((td64_07 + td64_10 - td64_01 + 0x7aaaaaaaaaaaaaa3ull) ^ p512[14]); \
	xl ^= p_23; \
	te += p_20 - p_06; \
	/* expand64_22(24); */ \
	p_24 = te + r64_1(p_09) + r64_2(p_11) + r64_3(p_13) \
		+ r64_4(p_15) + r64_5(p_17) + r64_6(p_19) + r64_7(p_21) \
		+ s64_4( p_22 ) + s64_5( p_23) \
		+ ((td64_08 + td64_11 - td64_02 + 0x7ffffffffffffff8ull) ^ p512[15]); \
	xh = xl ^ p_24; \
	to += p_21 - p_07; \
	/* expand64_21(25); */ \
	p_25 = to + r64_1(p_10) + r64_2(p_12) + r64_3(p_14) \
		+ r64_4(p_16) + r64_5(p_18) + r64_6(p_20) + r64_7(p_22) \
		+ s64_4( p_23) + s64_5( p_24) \
		+ ((td64_09 + td64_12 - td64_03 + 0x855555555555554dull) ^ p512[ 0]); \
	xh ^= p_25; \
	te += p_22 - p_08; \
	/* expand64_22(26); */ \
	p_26 = te + r64_1(p_11) + r64_2(p_13) + r64_3(p_15) \
		+ r64_4(p_17) + r64_5(p_19) + r64_6(p_21) + r64_7(p_23) \
		+ s64_4( p_24) + s64_5( p_25) \
		+ ((td64_10 + td64_13 - td64_04 + 0x8aaaaaaaaaaaaaa2ull) ^ p512[ 1]); \
	xh ^= p_26; \
	to += p_23 - p_09; \
	/* expand64_21(27); */ \
	p_27 = to + r64_1(p_12) + r64_2(p_14) + r64_3(p_16) \
		+ r64_4(p_18) + r64_5(p_20) + r64_6(p_22) + r64_7(p_24) \
		+ s64_4( p_25) + s64_5( p_26) \
		+ ((td64_11 + td64_14 - td64_05 + 0x8ffffffffffffff7ull) ^ p512[ 2]); \
	xh ^= p_27; \
	te += p_24 - p_10; \
	/* expand64_22(28); */ \
	p_28 = te + r64_1(p_13) + r64_2(p_15) + r64_3(p_17) \
		+ r64_4(p_19) + r64_5(p_21) + r64_6(p_23) + r64_7(p_25) \
		+ s64_4( p_26) + s64_5( p_27) \
		+ ((td64_12 + td64_15 - td64_06 + 0x955555555555554cull) ^ p512[ 3]); \
	xh ^= p_28; \
	to += p_25 - p_11; \
	/* expand64_21(29); */ \
	p_29 = to + r64_1(p_14) + r64_2(p_16) + r64_3(p_18) \
		+ r64_4(p_20) + r64_5(p_22) + r64_6(p_24) + r64_7(p_26) \
		+ s64_4( p_27) + s64_5( p_28) \
		+ ((td64_13 + td64_00 - td64_07 + 0x9aaaaaaaaaaaaaa1ull) ^ p512[ 4]); \
	xh ^= p_29; \
	te += p_26 - p_12; \
	/* expand64_22(30); */ \
	p_30  = te + r64_1(p_15) + r64_2(p_17) + r64_3(p_19) \
		+ r64_4(p_21) + r64_5(p_23) + r64_6(p_25) + r64_7(p_27) \
		+ s64_4( p_28) + s64_5( p_29) \
		+ ((td64_14 + td64_01 - td64_08 + 0x9ffffffffffffff6ull) ^ p512[ 5]); \
	xh ^= p_30; \
	to += p_27; to -=p_13; \
	/* expand64_21(31); */ \
	p_31 = to + r64_1(p_16) + r64_2(p_18) + r64_3(p_20) \
		+ r64_4(p_22) + r64_5(p_24) + r64_6(p_26) + r64_7(p_28) \
		+ s64_4( p_29) + s64_5( p_30) \
		+ ((td64_15 + td64_02 - td64_09 + 0xa55555555555554bull) ^ p512[ 6]); \
	xh ^= p_31; \
\
	/* Compute the double chaining pipe for the next message block */ \
	p512[0] = (shl(xh, 5) ^ shr(p_16,5) ^ d64( 0)) + (xl ^ p_24 ^ p_00); \
	p512[1] = (shr(xh, 7) ^ shl(p_17,8) ^ d64( 1)) + (xl ^ p_25 ^ p_01); \
	p512[2] = (shr(xh, 5) ^ shl(p_18,5) ^ d64( 2)) + (xl ^ p_26 ^ p_02); \
	p512[3] = (shr(xh, 1) ^ shl(p_19,5) ^ d64( 3)) + (xl ^ p_27 ^ p_03); \
	p512[4] = (shr(xh, 3) ^ p_20 ^ d64( 4)) + (xl ^ p_28 ^ p_04); \
	p512[5] = (shl(xh, 6) ^ shr(p_21,6) ^ d64( 5)) + (xl ^ p_29 ^ p_05); \
	p512[6] = (shr(xh, 4) ^ shl(p_22,6) ^ d64( 6)) + (xl ^ p_30 ^ p_06); \
	p512[7] = (shr(xh,11) ^ shl(p_23,2) ^ d64( 7)) + (xl ^ p_31 ^ p_07); \
\
	p512[ 8] = rotl64(p512[4], 9) + (xh ^ p_24 ^ d64( 8)) \
		+ (shl(xl,8) ^ p_23 ^ p_08); \
	p512[ 9] = rotl64(p512[5],10) + (xh ^ p_25 ^ d64( 9)) \
		+ (shr(xl,6) ^ p_16 ^ p_09); \
	p512[10] = rotl64(p512[6],11) + (xh ^ p_26 ^ d64(10)) \
		+ (shl(xl,6) ^ p_17 ^ p_10); \
	p512[11] = rotl64(p512[7],12) + (xh ^ p_27 ^ d64(11)) \
		+ (shl(xl,4) ^ p_18 ^ p_11); \
	p512[12] = rotl64(p512[0],13) + (xh ^ p_28 ^ d64(12)) \
		+ (shr(xl,3) ^ p_19 ^ p_12); \
	p512[13] = rotl64(p512[1],14) + (xh ^ p_29 ^ d64(13)) \
		+ (shr(xl,4) ^ p_20 ^ p_13); \
	p512[14] = rotl64(p512[2],15) + (xh ^ p_30 ^ d64(14)) \
		+ (shr(xl,7) ^ p_21 ^ p_14); \
	p512[15] = rotl64(p512[3],16) + (xh ^ p_31 ^ d64(15)) \
		+ (shr(xl,2) ^ p_22 ^ p_15); \
}


#define FinalCompression512()\
{\
	uint64_t p_00, p_01, p_02, p_03, p_04, p_05, p_06, p_07, p_08, p_09, \
		p_10, p_11, p_12, p_13, p_14, p_15, p_16, p_17, p_18, p_19, p_20, \
		p_21, p_22, p_23, p_24, p_25, p_26, p_27, p_28, p_29, p_30, p_31; \
	uint64_t t, te, to, xl, xh, *p = p512; \
\
	/* Mix the message block with the previous double pipe */ \
	p_00 = *p++ ^ 0xaaaaaaaaaaaaaaa0ull; \
	p_01 = *p++ ^ 0xaaaaaaaaaaaaaaa1ull; \
	p_02 = *p++ ^ 0xaaaaaaaaaaaaaaa2ull; \
	p_03 = *p++ ^ 0xaaaaaaaaaaaaaaa3ull; \
	p_04 = *p++ ^ 0xaaaaaaaaaaaaaaa4ull; \
	p_05 = *p++ ^ 0xaaaaaaaaaaaaaaa5ull; \
	p_06 = *p++ ^ 0xaaaaaaaaaaaaaaa6ull; \
	p_07 = *p++ ^ 0xaaaaaaaaaaaaaaa7ull; \
	p_08 = *p++ ^ 0xaaaaaaaaaaaaaaa8ull; \
	p_09 = *p++ ^ 0xaaaaaaaaaaaaaaa9ull; \
	p_10 = *p++ ^ 0xaaaaaaaaaaaaaaaaull; \
	p_11 = *p++ ^ 0xaaaaaaaaaaaaaaabull; \
	p_12 = *p++ ^ 0xaaaaaaaaaaaaaaacull; \
	p_13 = *p++ ^ 0xaaaaaaaaaaaaaaadull; \
	p_14 = *p++ ^ 0xaaaaaaaaaaaaaaaeull; \
	p_15 = *p++ ^ 0xaaaaaaaaaaaaaaafull; \
\
	/* These variables are used for the tweak */ \
	td64_00 = rotl64(p512[ 0], 1); \
	td64_01 = rotl64(p512[ 1], 2); \
	td64_02 = rotl64(p512[ 2], 3); \
	td64_03 = rotl64(p512[ 3], 4); \
	td64_04 = rotl64(p512[ 4], 5); \
	td64_05 = rotl64(p512[ 5], 6); \
	td64_06 = rotl64(p512[ 6], 7); \
	td64_07 = rotl64(p512[ 7], 8); \
	td64_08 = rotl64(p512[ 8], 9); \
	td64_09 = rotl64(p512[ 9],10); \
	td64_10 = rotl64(p512[10],11); \
	td64_11 = rotl64(p512[11],12); \
	td64_12 = rotl64(p512[12],13); \
	td64_13 = rotl64(p512[13],14); \
	td64_14 = rotl64(p512[14],15); \
	td64_15 = rotl64(p512[15],16); \
\
	{ \
		uint64_t t, t_16, t_17, t_18, t_19, t_20, t_21; \
\
		t_16 = p_01 - p_14; \
		t_17 = p_15 - p_12; \
		t_18 = p_14 - p_07; \
		t_19 = p_13 - p_06; \
		t_20 = p_08 - p_05; \
		t_21 = p_08 - p_01; \
\
		t = (p_05 + t_18 + p_10 + p_13); \
		p_16 = s64_0(t) + 0xaaaaaaaaaaaaaaa1ull; \
		t = (p_06 - p_08 + p_11 + p_14 - p_15); \
		p_17 = s64_1(t) + 0xaaaaaaaaaaaaaaa2ull; \
		t = (p_00 + p_07 + p_09 + t_17); \
		p_18 = s64_2(t) + 0xaaaaaaaaaaaaaaa3ull; \
		t = (p_00 + t_21 - p_10 + p_13); \
		p_19 = s64_3(t) + 0xaaaaaaaaaaaaaaa4ull; \
		t = (t_16 + p_02 + p_09 - p_11); \
		p_20 = s64_4(t) + 0xaaaaaaaaaaaaaaa5ull; \
		t = (p_03 - p_02 + p_10 + t_17); \
		p_21 = s64_0(t) + 0xaaaaaaaaaaaaaaa6ull; \
		t = (p_04 - p_00 - p_03 - p_11+p_13); \
		p_22 = s64_1(t) + 0xaaaaaaaaaaaaaaa7ull; \
		t = (t_16 - p_04 - p_05 - p_12); \
		p_23 = s64_2(t) + 0xaaaaaaaaaaaaaaa8ull; \
		t = (p_02 - p_05 + t_19 - p_15); \
		p_24 = s64_3(t) + 0xaaaaaaaaaaaaaaa9ull; \
		t = (p_00 - p_03 + p_06 + t_18); \
		p_25 = s64_4(t) + 0xaaaaaaaaaaaaaaaaull; \
		t = (t_21 - p_04 - p_07 + p_15); \
		p_26 = s64_0(t) + 0xaaaaaaaaaaaaaaabull; \
		t = (t_20 - p_00 - p_02 + p_09); \
		p_27 = s64_1(t) + 0xaaaaaaaaaaaaaaacull; \
		t = (p_01 + p_03 - p_06 - p_09 + p_10); \
		p_28 = s64_2(t) + 0xaaaaaaaaaaaaaaadull; \
		t = (p_02 + p_04 + p_07 + p_10 + p_11); \
		p_29 = s64_3(t) + 0xaaaaaaaaaaaaaaaeull; \
		p_30 = (p_03 + t_20 - p_11 - p_12); \
		p_14 = s64_4(p_30) + 0xaaaaaaaaaaaaaaafull; \
		p_31 = (p_12 - p_04 + t_19 - p_09); \
		p_15 = s64_0(p_31) + 0xaaaaaaaaaaaaaaa0ull; \
	} \
\
	p_00 = p_16; \
	p_01 = p_17; \
	p_02 = p_18; \
	p_03 = p_19; \
	p_04 = p_20; \
	p_05 = p_21; \
	p_06 = p_22; \
	p_07 = p_23; \
	p_08 = p_24; \
	p_09 = p_25; \
	p_10 = p_26; \
	p_11 = p_27; \
	p_12 = p_28; \
	p_13 = p_29; \
\
	/*** The 16-round message expansion part ***/ \
\
	xl = p_16 = s64_1(p_00) + s64_2(p_01) + s64_3(p_02) \
		+ s64_0(p_03) + s64_1(p_04) + s64_2(p_05) + s64_3(p_06) \
		+ s64_0(p_07) + s64_1(p_08) + s64_2(p_09) + s64_3(p_10) \
		+ s64_0(p_11) + s64_1(p_12) + s64_2(p_13) + s64_3(p_14) \
		+ s64_0(p_15) + ((td64_00 + td64_03 - td64_10 \
		+ 0x5555555555555550ull) ^ 0xaaaaaaaaaaaaaaa7ull); \
	p_17 = s64_1(p_01) + s64_2(p_02) + s64_3(p_03) \
		+ s64_0(p_04) + s64_1(p_05) + s64_2(p_06) + s64_3(p_07) \
		+ s64_0(p_08) + s64_1(p_09) + s64_2(p_10) + s64_3(p_11) \
		+ s64_0(p_12) + s64_1(p_13) + s64_2(p_14) + s64_3(p_15) \
		+ s64_0(p_16) + ((td64_01 + td64_04 - td64_11 \
		+ 0x5aaaaaaaaaaaaaa5ull) ^ 0xaaaaaaaaaaaaaaa8ull); \
	xl ^= p_17; \
	te = p_14 + p_12 + p_10 + p_08 + p_06 + p_04 + p_02; \
	to = p_15 + p_13 + p_11 + p_09 + p_07 + p_05 + p_03; \
\
	/* expand64_22(18); */ \
	p_18 = te + r64_1(p_03) + r64_2(p_05) + r64_3(p_07) \
		+ r64_4(p_09) + r64_5(p_11) + r64_6(p_13) + r64_7(p_15) \
		+ s64_4(p_16) + s64_5( p_17) + ((td64_02 + td64_05 - td64_12 \
		+ 0x5ffffffffffffffaull) ^ 0xaaaaaaaaaaaaaaa9ull); \
	xl ^= p_18; \
	/* expand64_21(19); */ \
	p_19 = to + r64_1(p_04) + r64_2(p_06) + r64_3(p_08) \
		+ r64_4(p_10) + r64_5(p_12) + r64_6(p_14) + r64_7(p_16) \
		+ s64_4(p_17) + s64_5( p_18) + ((td64_03 + td64_06 - td64_13 \
		+ 0x655555555555554full) ^ 0xaaaaaaaaaaaaaaaaull); \
	xl ^= p_19; \
	te += p_16 - p_02; \
	/* expand64_22(20); */ \
	p_20 = te + r64_1(p_05) + r64_2(p_07) + r64_3(p_09) \
		+ r64_4(p_11) + r64_5(p_13) + r64_6(p_15) + r64_7(p_17) \
		+ s64_4(p_18) + s64_5( p_19) + ((td64_04 + td64_07 - td64_14 \
		+ 0x6aaaaaaaaaaaaaa4ull) ^ 0xaaaaaaaaaaaaaaabull); \
	xl ^= p_20; \
	to += p_17 - p_03; \
	/* expand64_21(21); */ \
	p_21 = to + r64_1(p_06) + r64_2(p_08) + r64_3(p_10) \
		+ r64_4(p_12) + r64_5(p_14) + r64_6(p_16) + r64_7(p_18) \
		+ s64_4(p_19) + s64_5( p_20) + ((td64_05 + td64_08 - td64_15 \
		+ 0x6ffffffffffffff9ull) ^ 0xaaaaaaaaaaaaaaacull); \
	xl ^= p_21; \
	te += p_18 - p_04; \
	/* expand64_22(22); */ \
	p_22  = te + r64_1(p_07) + r64_2(p_09) + r64_3(p_11) \
		+ r64_4(p_13) + r64_5(p_15) + r64_6(p_17) + r64_7(p_19) \
		+ s64_4(p_20) + s64_5( p_21) + ((td64_06 + td64_09 - td64_00 \
		+ 0x755555555555554eull) ^ 0xaaaaaaaaaaaaaaadull); \
	xl ^= p_22; \
	to += p_19 - p_05; \
	/* expand64_21(23); */ \
	p_23 = to + r64_1(p_08) + r64_2(p_10) + r64_3(p_12) \
		+ r64_4(p_14) + r64_5(p_16) + r64_6(p_18) + r64_7(p_20) \
		+ s64_4(p_21) + s64_5( p_22) + ((td64_07 + td64_10 - td64_01 \
		+ 0x7aaaaaaaaaaaaaa3ull) ^ 0xaaaaaaaaaaaaaaaeull); \
	xl ^= p_23; \
	te += p_20 - p_06; \
	/* expand64_22(24); */ \
	p_24 = te + r64_1(p_09) + r64_2(p_11) + r64_3(p_13) \
		+ r64_4(p_15) + r64_5(p_17) + r64_6(p_19) + r64_7(p_21) \
		+ s64_4(p_22) + s64_5( p_23) + ((td64_08 + td64_11 - td64_02\
		 + 0x7ffffffffffffff8ull) ^ 0xaaaaaaaaaaaaaaafull); \
	xh = xl^p_24; \
	to += p_21 - p_07; \
	/* expand64_21(25); */ \
	p_25 = to + r64_1(p_10) + r64_2(p_12) + r64_3(p_14) \
		+ r64_4(p_16) + r64_5(p_18) + r64_6(p_20) + r64_7(p_22) \
		+ s64_4(p_23) + s64_5( p_24) + ((td64_09 + td64_12 - td64_03 \
		+ 0x855555555555554dull) ^ 0xaaaaaaaaaaaaaaa0ull); \
	xh ^= p_25; \
	te += p_22 - p_08; \
	/* expand64_22(26); */ \
	p_26 = te + r64_1(p_11) + r64_2(p_13) + r64_3(p_15) \
		+ r64_4(p_17) + r64_5(p_19) + r64_6(p_21) + r64_7(p_23) \
		+ s64_4(p_24) + s64_5( p_25) + ((td64_10 + td64_13 - td64_04 \
		+ 0x8aaaaaaaaaaaaaa2ull) ^ 0xaaaaaaaaaaaaaaa1ull); \
	xh ^= p_26; \
	to += p_23 - p_09; \
	/* expand64_21(27); */ \
	p_27 = to + r64_1(p_12) + r64_2(p_14) + r64_3(p_16) \
		+ r64_4(p_18) + r64_5(p_20) + r64_6(p_22) + r64_7(p_24) \
		+ s64_4(p_25) + s64_5( p_26) + ((td64_11 + td64_14 - td64_05 \
		+ 0x8ffffffffffffff7ull) ^ 0xaaaaaaaaaaaaaaa2ull); \
	xh ^= p_27; \
	te += p_24 - p_10; \
	/* expand64_22(28); */ \
	p_28 = te + r64_1(p_13) + r64_2(p_15) + r64_3(p_17) \
		+ r64_4(p_19) + r64_5(p_21) + r64_6(p_23) + r64_7(p_25) \
		+ s64_4( p_26) + s64_5( p_27) + ((td64_12 + td64_15 - td64_06 \
		+ 0x955555555555554cull) ^ 0xaaaaaaaaaaaaaaa3ull); \
	xh ^= p_28; \
	to += p_25 - p_11; \
	/* expand64_21(29); */ \
	p_29 = to + r64_1(p_14) + r64_2(p_16) + r64_3(p_18) \
		+ r64_4(p_20) + r64_5(p_22) + r64_6(p_24) + r64_7(p_26) \
		+ s64_4(p_27) + s64_5( p_28) + ((td64_13 + td64_00 - td64_07 \
		+ 0x9aaaaaaaaaaaaaa1ull) ^ 0xaaaaaaaaaaaaaaa4ull); \
	xh ^= p_29; \
	te += p_26 - p_12; \
	/* expand64_22(30); */ \
	p_30  = te + r64_1(p_15) + r64_2(p_17) + r64_3(p_19) \
		+ r64_4(p_21) + r64_5(p_23) + r64_6(p_25) + r64_7(p_27) \
		+ s64_4(p_28) + s64_5( p_29) + ((td64_14 + td64_01 - td64_08 \
		+ 0x9ffffffffffffff6ull) ^ 0xaaaaaaaaaaaaaaa5ull); \
	xh ^= p_30; \
	to += p_27; to -=p_13; \
	/* expand64_21(31); */ \
	p_31 = to + r64_1(p_16) + r64_2(p_18) + r64_3(p_20) \
		+ r64_4(p_22) + r64_5(p_24) + r64_6(p_26) + r64_7(p_28) \
		+ s64_4(p_29) + s64_5( p_30) + ((td64_15 + td64_02 - td64_09 \
		+ 0xa55555555555554bull) ^ 0xaaaaaaaaaaaaaaa6ull); \
	xh ^= p_31; \
\
	/* Compute the double chaining pipe for the final message */ \
	p512[0] = (shl(xh, 5) ^ shr(p_16,5) ^ p512[ 0]) + (xl ^ p_24 ^ p_00); \
	p512[1] = (shr(xh, 7) ^ shl(p_17,8) ^ p512[ 1]) + (xl ^ p_25 ^ p_01); \
	p512[2] = (shr(xh, 5) ^ shl(p_18,5) ^ p512[ 2]) + (xl ^ p_26 ^ p_02); \
	p512[3] = (shr(xh, 1) ^ shl(p_19,5) ^ p512[ 3]) + (xl ^ p_27 ^ p_03); \
	p512[4] = (shr(xh, 3) ^ p_20 ^ p512[ 4]) + (xl ^ p_28 ^ p_04); \
	p512[5] = (shl(xh, 6) ^ shr(p_21,6) ^ p512[ 5]) + (xl ^ p_29 ^ p_05); \
	p512[6] = (shr(xh, 4) ^ shl(p_22,6) ^ p512[ 6]) + (xl ^ p_30 ^ p_06); \
	p512[7] = (shr(xh,11) ^ shl(p_23,2) ^ p512[ 7]) + (xl ^ p_31 ^ p_07); \
\
	p512[ 8] = rotl64(p512[4], 9) + (xh ^ p_24 ^ p512[ 8]) \
		+ (shl(xl,8) ^ p_23 ^ p_08); \
	p512[ 9] = rotl64(p512[5],10) + (xh ^ p_25 ^ p512[ 9]) \
		+ (shr(xl,6) ^ p_16 ^ p_09); \
	p512[10] = rotl64(p512[6],11) + (xh ^ p_26 ^ p512[10]) \
		+ (shl(xl,6) ^ p_17 ^ p_10); \
	p512[11] = rotl64(p512[7],12) + (xh ^ p_27 ^ p512[11]) \
		+ (shl(xl,4) ^ p_18 ^ p_11); \
	p512[12] = rotl64(p512[0],13) + (xh ^ p_28 ^ p512[12]) \
		+ (shr(xl,3) ^ p_19 ^ p_12); \
	p512[13] = rotl64(p512[1],14) + (xh ^ p_29 ^ p512[13]) \
		+ (shr(xl,4) ^ p_20 ^ p_13); \
	p512[14] = rotl64(p512[2],15) + (xh ^ p_30 ^ p512[14]) \
		+ (shr(xl,7) ^ p_21 ^ p_14); \
	p512[15] = rotl64(p512[3],16) + (xh ^ p_31 ^ p512[15]) \
		+ (shr(xl,2) ^ p_22 ^ p_15); \
}


HashReturn
Init(hashState *state, int hashbitlen)
{
	switch(hashbitlen)
	{
		case 224:
		state->hashbitlen = 224;
		// #1 Between comments #1 and #2 add algorithm specific initialization
		state->bits_processed = 0;
		state->unprocessed_bits = 0;
		memcpy(hashState224(state)->DoublePipe, i224p2,  16 * sizeof(uint32_t));
		// #2 Between comments #1 and #2 add algorithm specific initialization
		return SUCCESS;

		case 256:
		state->hashbitlen = 256;
		// #1 Between comments #1 and #2 add algorithm specific initialization
		state->bits_processed = 0;
		state->unprocessed_bits = 0;
		memcpy(hashState256(state)->DoublePipe, i256p2,  16 * sizeof(uint32_t));
		// #2 Between comments #1 and #2 add algorithm specific initialization
		return SUCCESS;

		case 384:		
		state->hashbitlen = 384;
		// #1 Between comments #1 and #2 add algorithm specific initialization
		state->bits_processed = 0;
		state->unprocessed_bits = 0;
		memcpy(hashState384(state)->DoublePipe, i384p2,  16 * sizeof(uint64_t));
		// #2 Between comments #1 and #2 add algorithm specific initialization
		return SUCCESS;

		case 512:
		state->hashbitlen = 512;
		// #1 Between comments #1 and #2 add algorithm specific initialization
		state->bits_processed = 0;
		state->unprocessed_bits = 0;
		memcpy(hashState224(state)->DoublePipe, i512p2,  16 * sizeof(uint64_t));
		// #2 Between comments #1 and #2 add algorithm specific initialization
		return SUCCESS;

		default:
			return BAD_HASHLEN;
	}
}


HashReturn
Update(hashState *state, const BitSequence *data, DataLength databitlen)
{
	uint32_t *data32, *p256;
	uint32_t td32_00, td32_01, td32_02, td32_03, td32_04, td32_05, td32_06,
		td32_07, td32_08, td32_09, td32_10, td32_11, td32_12, td32_13, td32_14,
		td32_15;
	uint64_t *data64, *p512;
	uint64_t td64_00, td64_01, td64_02, td64_03, td64_04, td64_05, td64_06,
		td64_07, td64_08, td64_09, td64_10, td64_11, td64_12, td64_13, td64_14,
		td64_15;
	int LastBytes;
#if defined(MACHINE_IS_BIG_ENDIAN)
	uint64_t swap64[16];
	uint32_t swap32[16];
	register int i;
#endif


	switch(state->hashbitlen) {
		case 224:
		case 256:
			if (state->unprocessed_bits > 0) {
				if (state->unprocessed_bits + databitlen >
					BlueMidnightWish256_BLOCK_SIZE * 8) {
					return BAD_CONSECUTIVE_CALL_TO_UPDATE;
				} else {
					LastBytes = (int)databitlen >> 3;
					memcpy(hashState256(state)->LastPart +
						(state->unprocessed_bits >> 3), data, LastBytes);
					state->unprocessed_bits += (int)databitlen;
					databitlen = state->unprocessed_bits;
					data32 = (uint32_t *)hashState256(state)->LastPart;
				}
			} else 
				data32 = (uint32_t *)data;

			p256   = hashState256(state)->DoublePipe;
			while (databitlen >= BlueMidnightWish256_BLOCK_SIZE * 8) {
				/* #1 between comments #1 and #2 add algorithm specifics */
#if defined(MACHINE_IS_BIG_ENDIAN)
				for (i = 0; i < 16; i++)
					ld_swap32(data32+i, swap32[i]);
#define d32(j) swap32[j]
#else
#define d32(j) data32[j]
#endif /*MACHINE_IS_BIG_ENDIAN*/
				Compression256();
#undef d32
				databitlen -= BlueMidnightWish256_BLOCK_SIZE * 8;
				state->bits_processed += BlueMidnightWish256_BLOCK_SIZE * 8;
				data32 += 16;
			}
			state->unprocessed_bits = (int)databitlen;
			if (databitlen > 0) {
				/* LastBytes = ceil(databitlen / 8) */
				LastBytes = ((~(((- (int)databitlen)>>3) & 0x1ff)) + 1) & 0x1ff;
				memcpy(hashState256(state)->LastPart, data32, LastBytes);
			}
			/* #2 between comments #1 and #2 add algorithm specifics */
			return SUCCESS;

		case 384:
		case 512:
			if (state->unprocessed_bits > 0) {
				if (state->unprocessed_bits + databitlen >
					BlueMidnightWish512_BLOCK_SIZE * 8) {
					return BAD_CONSECUTIVE_CALL_TO_UPDATE;
				} else {
					LastBytes = (int)databitlen >> 3;
					memcpy(hashState512(state)->LastPart +
						(state->unprocessed_bits >> 3), data, LastBytes);
					state->unprocessed_bits += (int)databitlen;
					databitlen = state->unprocessed_bits;
					data64 = (uint64_t *)hashState512(state)->LastPart;
				}
			}
			else 
				data64 = (uint64_t *)data;

			p512   = hashState512(state)->DoublePipe;
			while (databitlen >= BlueMidnightWish512_BLOCK_SIZE * 8) {
				/* #1 between comments #1 and #2 add algorithm specifics */
#if defined(MACHINE_IS_BIG_ENDIAN)
				for (i = 0; i < 16; i++)
					ld_swap64(data64+i, swap64[i]);
#define d64(j) swap64[j]
#define D64 swap64
#else
#define d64(j) data64[j]
#define D64 data64
#endif /*MACHINE_IS_BIG_ENDIAN*/
				Compression512();
#undef d64
#undef D64
				databitlen -= BlueMidnightWish512_BLOCK_SIZE * 8;
				state->bits_processed += BlueMidnightWish512_BLOCK_SIZE * 8;
				data64 += 16;
			}
			state->unprocessed_bits = (int)databitlen;
			if (databitlen > 0) {
				/* LastBytes = ceil(databitlen / 8) */
				LastBytes = ((~(((- (int)databitlen)>>3) & 0x3ff)) + 1) & 0x3ff;
				memcpy(hashState512(state)->LastPart, data64, LastBytes);
			}
			/* #2 between comments #1 and #2 add algorithm specifics */
			return SUCCESS;
		
		default:
			return BAD_HASHLEN; /* this should never happen */
	}
}


HashReturn
Final(hashState *state, BitSequence *hashval)
{
	uint32_t *data32, *p256;
	uint32_t td32_00, td32_01, td32_02, td32_03, td32_04, td32_05, td32_06,
		td32_07, td32_08, td32_09, td32_10, td32_11, td32_12, td32_13, td32_14,
		td32_15;
	uint64_t *data64, *p512, num_bits;
	uint64_t td64_00, td64_01, td64_02, td64_03, td64_04, td64_05, td64_06,
		td64_07, td64_08, td64_09, td64_10, td64_11, td64_12, td64_13, td64_14,
		td64_15;
	DataLength databitlen;
	int LastByte, PadOnePosition;
#if defined(MACHINE_IS_BIG_ENDIAN)
	uint64_t swap64[16];
	uint32_t swap32[16];
	register int i;
#endif


	num_bits = state->bits_processed + state->unprocessed_bits;

	switch(state->hashbitlen) {
		case 224:
		case 256:
			LastByte = (int)state->unprocessed_bits >> 3;
			PadOnePosition = 7 - (state->unprocessed_bits & 0x7);
			hashState256(state)->LastPart[LastByte] =
				(hashState256(state)->LastPart[LastByte]
				& (0xff << (PadOnePosition + 1))) ^ (0x1 << PadOnePosition);
			data64 = (uint64_t *)hashState256(state)->LastPart;

			if (state->unprocessed_bits < 448) {
				memset((hashState256(state)->LastPart) + LastByte + 1, 0x0,
					BlueMidnightWish256_BLOCK_SIZE - LastByte - 9 );
				databitlen = BlueMidnightWish256_BLOCK_SIZE * 8;
				st_swap64(num_bits, data64+7);
			} else {
				memset((hashState256(state)->LastPart) + LastByte + 1, 0x0,
					BlueMidnightWish256_BLOCK_SIZE * 2 - LastByte - 9 );
				databitlen = BlueMidnightWish256_BLOCK_SIZE * 16;
				st_swap64(num_bits, data64+15);
			}

			data32   = (uint32_t *)hashState256(state)->LastPart;
			p256     = hashState256(state)->DoublePipe;
			while (databitlen >= BlueMidnightWish256_BLOCK_SIZE * 8) {
				/* #1 between comments #1 and #2 add algorithm specifics */
#if defined(MACHINE_IS_BIG_ENDIAN)
				for (i = 0; i < 16; i++)
					ld_swap32(data32+i, swap32[i]);
#define d32(j) swap32[j]
#else
#define d32(j) data32[j]
#endif /*MACHINE_IS_BIG_ENDIAN*/
				Compression256();
#undef d32
				databitlen -= BlueMidnightWish256_BLOCK_SIZE * 8;
				data32 += 16;
			}
			/* #2 between comments #1 and #2 add algorithm specifics */
			break;

		case 384:
		case 512:
			LastByte = (int)state->unprocessed_bits >> 3;
			PadOnePosition = 7 - (state->unprocessed_bits & 0x07);
			hashState512(state)->LastPart[LastByte] =
				(hashState512(state)->LastPart[LastByte]
				& (0xff << (PadOnePosition + 1))) ^ (0x1 << PadOnePosition);
			data64 = (uint64_t *)hashState512(state)->LastPart;

			if (state->unprocessed_bits < 960) {
				memset((hashState512(state)->LastPart) + LastByte + 1, 0x0,
					BlueMidnightWish512_BLOCK_SIZE - LastByte - 9 );
				databitlen = BlueMidnightWish512_BLOCK_SIZE * 8;
				st_swap64(num_bits, data64+15);
			} else {
				memset((hashState512(state)->LastPart) + LastByte + 1, 0x0,
					BlueMidnightWish512_BLOCK_SIZE * 2 - LastByte - 9 );
				databitlen = BlueMidnightWish512_BLOCK_SIZE * 16;
				st_swap64(num_bits, data64+31);
			}

			p512   = hashState512(state)->DoublePipe;
			while (databitlen >= BlueMidnightWish512_BLOCK_SIZE * 8) {
				/* #1 between comments #1 and #2 add algorithm specifics */
#if defined(MACHINE_IS_BIG_ENDIAN)
				for (i = 0; i < 16; i++)
					ld_swap64(data64+i, swap64[i]);
#define d64(j) swap64[j]
#define D64 swap64
#else
#define d64(j) data64[j]
#define D64 data64
#endif /*MACHINE_IS_BIG_ENDIAN*/
				Compression512();
#undef d64
#undef D64
				databitlen -= BlueMidnightWish512_BLOCK_SIZE * 8;
				data64 += 16;
			}
			break;
			/* #2 between comments #1 and #2 add algorithm specifics */
		
		default:
			return BAD_HASHLEN; /* this should never happen */
	}

	/*
	 * The FinalCompressionXXX() macros implement the tweak for Blue Midnight
	 * Wish as submitted on 2009-09-15. Here, the role of the message has the
	 * obtained final double pipe, and the role of the initial double pipe is
	 * a constant.
	 */
	switch(state->hashbitlen) {
		case 224:
		case 256:
			FinalCompression256();
			break;

		case 384:
		case 512:
			FinalCompression512();
			break;
		
		default:
			return BAD_HASHLEN; /* this should never happen */
	}

	switch(state->hashbitlen) {
		case 224:
#if defined(MACHINE_IS_BIG_ENDIAN)
			{
				uint32_t *d32 = (uint32_t *)hashval;
				uint32_t *s32 = p256 + 9;
				register int j;

				for (j = 0; j < BlueMidnightWish224_DIGEST_SIZE>>2; j++)
					st_swap32(s32[j], d32+j);
			}
#else
			memcpy(hashval, p256 + 9, BlueMidnightWish224_DIGEST_SIZE );
#endif
			return SUCCESS;
		case 256:
#if defined(MACHINE_IS_BIG_ENDIAN)
			{
				uint32_t *d32 = (uint32_t *)hashval;
				uint32_t *s32 = p256 + 8;
				register int j;

				for (j = 0; j < BlueMidnightWish256_DIGEST_SIZE>>2; j++)
					st_swap32(s32[j], d32+j);
			}
#else
			memcpy(hashval, p256 + 8, BlueMidnightWish256_DIGEST_SIZE );
#endif
			return SUCCESS;
		case 384:
#if defined(MACHINE_IS_BIG_ENDIAN)
			{
				uint64_t *d64 = (uint64_t *)hashval;
				uint64_t *s64 = p512 + 10;
				register int j;

				for (j = 0; j < BlueMidnightWish384_DIGEST_SIZE>>3; j++)
					st_swap64(s64[j], d64+j);
			}
#else
			memcpy(hashval, p512 + 10, BlueMidnightWish384_DIGEST_SIZE );
#endif
			return SUCCESS;
		case 512:
#if defined(MACHINE_IS_BIG_ENDIAN)
			{
				uint64_t *d64 = (uint64_t *)hashval;
				uint64_t *s64 = p512 + 8;
				register int j;

				for (j = 0; j < BlueMidnightWish512_DIGEST_SIZE>>3; j++)
					st_swap64(s64[j], d64+j);
			}
#else
			memcpy(hashval, p512 + 8,  BlueMidnightWish512_DIGEST_SIZE );
#endif
			return SUCCESS;
		default:
			return BAD_HASHLEN; /* this should never happen */
	}
}


HashReturn
Hash(int hashbitlen, const BitSequence *data, DataLength databitlen,
	 BitSequence *hashval)
{
	HashReturn qq;
	hashState state;

	qq = Init(&state, hashbitlen);
	if (qq != SUCCESS)
		return(qq);

	qq = Update(&state, data, databitlen);
	if (qq != SUCCESS)
		return(qq);

	return Final(&state, hashval);
}
